import autolens as al
import numpy as np
import pytest

from test.unit.mock.data.mock_grids import MockPixelizationGrid


class TestRectangular:
    def test__5_simple_grid__no_sub_grid(self):

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, False, True, True, True],
                [True, True, False, False, False, True, True],
                [True, True, True, False, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=1.0)

        # Source-plane comprises 5 grid, so 5 masked_image pixels traced to the pix-plane.

        grid = al.Grid(
            arr=np.array(
                [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
            ),
            mask=mask,
            sub_grid_size=1,
        )

        pixelization_grid = MockPixelizationGrid(arr=grid)

        # There is no sub-grid, so our grid are just the masked_image grid (note the NumPy weighted_data structure
        # ensures this has no sub-gridding)

        pix = al.pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid,
            pixelization_grid=pixelization_grid,
            inversion_uses_border=False,
            hyper_image=np.ones((2, 2)),
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)
        assert (mapper.hyper_image == np.ones((2, 2))).all()

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            pixel_neighbors=mapper.geometry.pixel_neighbors,
            pixel_neighbors_size=mapper.geometry.pixel_neighbors_size,
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__15_grid__no_sub_grid(self):

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, False, False, False, False, False, True],
                [True, False, False, False, False, False, True],
                [True, False, False, False, False, False, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=1.0)

        # There is no sub-grid, so our grid are just the masked_image grid (note the NumPy weighted_data structure
        # ensures this has no sub-gridding)
        grid = al.Grid(
            arr=np.array(
                [
                    [0.9, -0.9],
                    [1.0, -1.0],
                    [1.1, -1.1],
                    [0.9, 0.9],
                    [1.0, 1.0],
                    [1.1, 1.1],
                    [-0.01, 0.01],
                    [0.0, 0.0],
                    [0.01, 0.01],
                    [-0.9, -0.9],
                    [-1.0, -1.0],
                    [-1.1, -1.1],
                    [-0.9, 0.9],
                    [-1.0, 1.0],
                    [-1.1, 1.1],
                ]
            ),
            mask=mask,
            sub_grid_size=1,
        )

        pixelization_grid = MockPixelizationGrid(arr=grid)

        pix = al.pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=False
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.2, 2.2), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__5_simple_grid__include_sub_grid(self):

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, False, True, True, True],
                [True, True, False, False, False, True, True],
                [True, True, True, False, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=2.0)

        # Assume a 2x2 sub-grid, so each of our 5 masked_image-pixels are split into 4.
        # The grid below is unphysical in that the (0.0, 0.0) terms on the end of each sub-grid probably couldn't
        # happen for a real lensing calculation. This is to make a mapping_matrix matrix which explicitly tests the
        # sub-grid.
        grid = al.Grid(
            arr=np.array(
                [
                    [1.0, -1.0],
                    [1.0, -1.0],
                    [1.0, -1.0],
                    [1.0, 1.0],
                    [1.0, 1.0],
                    [1.0, 1.0],
                    [-1.0, -1.0],
                    [-1.0, -1.0],
                    [-1.0, -1.0],
                    [-1.0, 1.0],
                    [-1.0, 1.0],
                    [-1.0, 1.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                    [0.0, 0.0],
                ]
            ),
            mask=mask,
            sub_grid_size=2,
        )

        pixelization_grid = MockPixelizationGrid(arr=grid)

        pix = al.pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=False
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.75, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25, 0.0, 0.75],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__grid__requires_border_relocation(self):

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, False, True, True, True],
                [True, True, False, False, False, True, True],
                [True, True, True, False, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = al.Mask(array=mask, pixel_scale=1.0)

        grid = al.Grid(
            arr=np.array(
                [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [-1.0, -1.0]]
            ),
            mask=mask,
            sub_grid_size=1,
        )

        pixelization_grid = MockPixelizationGrid(grid)

        pix = al.pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=True
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()


class TestVoronoiMagnification:
    def test__3x3_simple_grid(self):

        mask = al.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        grid = np.array(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-1.0, -1.0],
                [-1.0, 0.0],
                [-1.0, 1.0],
            ]
        )

        grid = al.Grid(arr=grid, mask=mask)

        pix = al.pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=pix.shape, grid=grid.unlensed_unsubbed_1d
        )

        pixelization_grid = MockPixelizationGrid(
            arr=sparse_to_grid.sparse,
            mask_1d_index_to_nearest_pixelization_1d_index=sparse_to_grid.mask_1d_index_to_sparse_1d_index,
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid,
            pixelization_grid=pixelization_grid,
            inversion_uses_border=False,
            hyper_image=np.ones((2, 2)),
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)
        assert (mapper.hyper_image == np.ones((2, 2))).all()

        assert isinstance(mapper, al.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask(self):

        mask = al.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        grid = np.array([[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0], [-1.0, 0.0]])

        grid = al.Grid(arr=grid, mask=mask)

        pix = al.pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=pix.shape, grid=grid.unlensed_unsubbed_1d
        )

        pixelization_grid = MockPixelizationGrid(
            arr=grid,
            mask_1d_index_to_nearest_pixelization_1d_index=sparse_to_grid.mask_1d_index_to_sparse_1d_index,
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=False
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert isinstance(mapper, al.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask_and_sub_grid(self):

        mask = al.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        grid = np.array(
            [
                [1.01, 0.0],
                [1.01, 0.0],
                [1.01, 0.0],
                [0.01, 0.0],
                [0.0, -1.0],
                [0.0, -1.0],
                [0.0, -1.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.0, 1.01],
                [0.0, 1.01],
                [0.0, 1.01],
                [0.01, 0.0],
                [-1.01, 0.0],
                [-1.01, 0.0],
                [-1.01, 0.0],
                [0.01, 0.0],
            ]
        )

        grid = al.Grid(arr=grid, mask=mask, sub_grid_size=2)

        pix = al.pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=pix.shape, grid=grid.unlensed_unsubbed_1d
        )

        pixelization_grid = MockPixelizationGrid(
            arr=sparse_to_grid.sparse,
            mask_1d_index_to_nearest_pixelization_1d_index=sparse_to_grid.mask_1d_index_to_sparse_1d_index,
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=False
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.02, 2.01), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.005), 1.0e-4)

        assert isinstance(mapper, al.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.75, 0.0, 0.25, 0.0, 0.0],
                    [0.0, 0.75, 0.25, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.25, 0.75, 0.0],
                    [0.0, 0.0, 0.25, 0.0, 0.75],
                ]
            )
        ).all()

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask_with_offset_centre(self):

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, False, True],
                    [True, True, False, False, False],
                    [True, True, True, False, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
        )

        grid = np.array([[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]])

        grid = al.Grid(arr=grid, mask=mask, sub_grid_size=1)

        pix = al.pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_grid = al.SparseToGrid.from_grid_and_unmasked_2d_grid_shape(
            unmasked_sparse_shape=pix.shape, grid=grid.unlensed_unsubbed_1d
        )

        pixelization_grid = MockPixelizationGrid(
            arr=sparse_to_grid.sparse,
            mask_1d_index_to_nearest_pixelization_1d_index=sparse_to_grid.mask_1d_index_to_sparse_1d_index,
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=grid, pixelization_grid=pixelization_grid, inversion_uses_border=False
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((1.0, 1.0), 1.0e-4)

        assert isinstance(mapper, al.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = al.regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()



import os

import autofit as af
import autolens as al


# Welcome to the pipeline runner. This tool allows you to load strong lens data, and pass it to pipelines for a
# PyAutoLens analysis. To show you around, we'll load up some example instrument and run it through some of the example
# pipelines that come distributed with PyAutoLens.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out different lens
# names and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# It is convenient to specify the data type and data name as a string, so that if the pipeline is applied to multiple
# images we don't have to change all of the path entries in the load_ccd_data_from_fits function below.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + "../../outputs/PyAutoLens/"
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=output_path
)

data_name = "rjlens"
data_filter = "f814w"

pixel_scale = 0.04

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_name, data_filter]
)

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
)

mask = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.7
)

positions = al.load_positions(positions_path=data_path + "positions.dat")

data_name = "rjlens_hyper"

pipeline_settings = al.PipelineSettingsHyper(
    hyper_galaxies=True,
    hyper_background_noise=True,
    include_shear=True,
    fix_lens_light=False,
    align_bulge_disk_centre=False,
    align_bulge_disk_phi=False,
    align_bulge_disk_axis_ratio=False,
)

from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.initialize import (
    lens_bulge_disk_sie__source_sersic,
)

from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.inversion.from_initialize import (
    lens_bulge_disk_sie__source_inversion,
)

from workspace_jam.pipelines.hyper.with_lens_light.bulge_disk.power_law.from_inversion import (
    lens_bulge_disk_power_law__source_inversion,
)

pipeline_initialize = lens_bulge_disk_sie__source_sersic.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_name, data_filter],
    positions_threshold=1.0,
)

pipeline_inversion = lens_bulge_disk_sie__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_name, data_filter],
    positions_threshold=1.0,
)
pipeline_power_law = lens_bulge_disk_power_law__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_name, data_filter],
    positions_threshold=1.0,
)

pipeline = pipeline_initialize + pipeline_inversion + pipeline_power_law

pipeline.run(data=ccd_data, mask=mask, positions=positions)



import autolens as al
import numpy as np

grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
)

sie = al.mass_profiles.EllipticalIsothermalKormann(
    centre=(0.0, 0.0), einstein_radius=1.0, axis_ratio=0.2, phi=45.0
)
convergence_0 = sie.deflections_from_grid(grid=grid, return_in_2d=True)[:, :, 0]
deflections_y_0 = al.ScaledSquarePixelArray(array=convergence_0, pixel_scale=0.05)

sie = al.mass_profiles.EllipticalIsothermalKormann(
    centre=(0.0, 0.0), einstein_radius=1.0, axis_ratio=0.2, phi=135.0
)
convergence_1 = sie.deflections_from_grid_2(grid=grid, return_in_2d=True)[:, :, 0]
deflections_y_1 = al.ScaledSquarePixelArray(array=convergence_1, pixel_scale=0.05)

residuals = convergence_0 - convergence_1

print(np.max(residuals))
residuals = al.ScaledSquarePixelArray(array=residuals, pixel_scale=0.05)
al.array_plotters.plot_array(array=residuals)



def likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
    chi_squared, regularization_term, noise_normalization
):
    """Compute the likelihood of an inversion's fit to the datas, including a regularization term which \
    comes from an inversion:

    Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the observed datas.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    noise_normalization : float
        The normalization noise_map-term for the observed datas's noise-map.
    """
    return -0.5 * (chi_squared + regularization_term + noise_normalization)


def evidence_from_inversion_terms(
    chi_squared,
    regularization_term,
    log_curvature_regularization_term,
    log_regularization_term,
    noise_normalization,
):
    """Compute the evidence of an inversion's fit to the datas, where the evidence includes a number of \
    terms which quantify the complexity of an inversion's reconstruction (see the *inversion* module):

    Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Log(Covariance_Regularization_Term) -
                       Log(Regularization_Matrix_Term) + Noise_Term]

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the observed datas.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    log_curvature_regularization_term : float
        The log of the determinant of the sum of the curvature and regularization matrices.
    log_regularization_term : float
        The log of the determinant o the regularization matrix.
    noise_normalization : float
        The normalization noise_map-term for the observed datas's noise-map.
    """
    return -0.5 * (
        chi_squared
        + regularization_term
        + log_curvature_regularization_term
        - log_regularization_term
        + noise_normalization
    )



import autolens as al
import os

import numpy as np


test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestSubMask1DIndexToMask1DIndex(object):
    def test__3x3_mask_with_1_pixel__2x2_sub_grid__correct_sub_mask_1d_index_to_mask_1d_index(
        self
    ):
        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_mask_1d_index_to_mask_1d_index = al.mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask, sub_grid_size=2
        )

        assert (sub_mask_1d_index_to_mask_1d_index == np.array([0, 0, 0, 0])).all()

    def test__3x3_mask_with_row_of_pixels_pixel__2x2_sub_grid__correct_sub_mask_1d_index_to_mask_1d_index(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        sub_mask_1d_index_to_mask_1d_index = al.mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask, sub_grid_size=2
        )

        assert (
            sub_mask_1d_index_to_mask_1d_index
            == np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
        ).all()

    def test__3x3_mask_with_row_of_pixels_pixel__3x3_sub_grid__correct_sub_mask_1d_index_to_mask_1d_index(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        sub_mask_1d_index_to_mask_1d_index = al.mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask, sub_grid_size=3
        )

        assert (
            sub_mask_1d_index_to_mask_1d_index
            == np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                ]
            )
        ).all()


class TestMask1DIndexToSubMask1DIndexes(object):
    def test__3x3_mask_with_1_pixel__2x2_sub_grid__correct_mask_1d_index_to_sub_mask_1d_indexes(
        self
    ):

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        mask_1d_index_to_sub_mask_1d_indexes = al.mask_mapping_util.mask_1d_index_to_sub_mask_1d_indexes_from_mask(
            mask, sub_grid_size=2
        )

        assert mask_1d_index_to_sub_mask_1d_indexes == [[0, 1, 2, 3]]

    def test__3x3_mask_with_row_of_pixels_pixel__2x2_sub_grid__correct_mask_1d_index_to_sub_mask_1d_indexes(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        mask_1d_index_to_sub_mask_1d_indexes = al.mask_mapping_util.mask_1d_index_to_sub_mask_1d_indexes_from_mask(
            mask, sub_grid_size=2
        )

        assert mask_1d_index_to_sub_mask_1d_indexes == [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
        ]

    def test__3x3_mask_with_row_of_pixels_pixel__3x3_sub_grid__correct_mask_1d_index_to_sub_mask_1d_indexes(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        mask_1d_index_to_sub_mask_1d_indexes = al.mask_mapping_util.mask_1d_index_to_sub_mask_1d_indexes_from_mask(
            mask, sub_grid_size=3
        )

        assert mask_1d_index_to_sub_mask_1d_indexes == [
            [0, 1, 2, 3, 4, 5, 6, 7, 8],
            [9, 10, 11, 12, 13, 14, 15, 16, 17],
            [18, 19, 20, 21, 22, 23, 24, 25, 26],
        ]


class TestSubMask2dToSubMask1d(object):
    def test__mask_if_full_of_false__indexes_are_ascending_order(self):

        mask_2d = np.full(fill_value=False, shape=(3, 3))

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (sub_two_to_one == np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])).all()

        mask_2d = np.full(fill_value=False, shape=(2, 3))

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (sub_two_to_one == np.array([[0, 1, 2], [3, 4, 5]])).all()

        mask_2d = np.full(fill_value=False, shape=(3, 2))

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (sub_two_to_one == np.array([[0, 1], [2, 3], [4, 5]])).all()

    def test__mask_has_true_and_falses__minus_ones_in_place_of_trues_and_falses_count_correctly(
        self
    ):

        mask_2d = np.array(
            [[False, True, False], [True, True, False], [False, False, True]]
        )

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (sub_two_to_one == np.array([[0, -1, 1], [-1, -1, 2], [3, 4, -1]])).all()

        mask_2d = np.array(
            [
                [False, True, True, False],
                [True, True, False, False],
                [False, False, True, False],
            ]
        )

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (
            sub_two_to_one == np.array([[0, -1, -1, 1], [-1, -1, 2, 3], [4, 5, -1, 6]])
        ).all()

        mask_2d = np.array(
            [
                [False, True, False],
                [True, True, False],
                [False, False, True],
                [False, False, True],
            ]
        )

        sub_two_to_one = al.mask_mapping_util.sub_mask_2d_index_to_sub_mask_1d_index_from_sub_mask(
            sub_mask=mask_2d
        )

        assert (
            sub_two_to_one
            == np.array([[0, -1, 1], [-1, -1, 2], [3, 4, -1], [5, 6, -1]])
        ).all()


class TestSubOneToTwo(object):
    def test__simple_mappings__sub_grid_size_is_1(self):

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=1
        )

        assert (sub_one_to_two == np.array([[1, 1]])).all()

        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, True]]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=1
        )

        assert (
            sub_one_to_two == np.array([[0, 1], [1, 0], [1, 1], [1, 2], [2, 1]])
        ).all()

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=1
        )

        assert (
            sub_one_to_two == np.array([[0, 1], [1, 0], [1, 1], [1, 2], [2, 1], [2, 3]])
        ).all()

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, False],
            ]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=1
        )

        assert (
            sub_one_to_two == np.array([[0, 1], [1, 0], [1, 1], [1, 2], [2, 1], [3, 2]])
        ).all()

    def test__simple_grid_mappings__sub_grid_size_2(self):

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=2
        )

        assert (sub_one_to_two == np.array([[2, 2], [2, 3], [3, 2], [3, 3]])).all()

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=3
        )

        assert (
            sub_one_to_two
            == np.array(
                [[3, 3], [3, 4], [3, 5], [4, 3], [4, 4], [4, 5], [5, 3], [5, 4], [5, 5]]
            )
        ).all()

        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, True]]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=2
        )

        assert (
            sub_one_to_two
            == np.array(
                [
                    [0, 2],
                    [0, 3],
                    [1, 2],
                    [1, 3],
                    [2, 0],
                    [2, 1],
                    [3, 0],
                    [3, 1],
                    [2, 2],
                    [2, 3],
                    [3, 2],
                    [3, 3],
                    [2, 4],
                    [2, 5],
                    [3, 4],
                    [3, 5],
                    [4, 2],
                    [4, 3],
                    [5, 2],
                    [5, 3],
                ]
            )
        ).all()

        mask = np.array(
            [
                [True, True, True, True],
                [False, True, True, True],
                [True, False, True, True],
            ]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=2
        )

        assert (
            sub_one_to_two
            == np.array(
                [[2, 0], [2, 1], [3, 0], [3, 1], [4, 2], [4, 3], [5, 2], [5, 3]]
            )
        ).all()

        mask = np.array(
            [
                [True, True, True],
                [True, False, True],
                [True, True, True],
                [True, True, False],
            ]
        )

        sub_one_to_two = al.mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
            mask=mask, sub_grid_size=2
        )

        assert (
            sub_one_to_two
            == np.array(
                [[2, 2], [2, 3], [3, 2], [3, 3], [6, 4], [6, 5], [7, 4], [7, 5]]
            )
        ).all()



from autolens.model.profiles import mass_profiles as mp
from astropy import cosmology as cosmo

cosmology = cosmo.Planck15
redshift_lens = 0.6
redshift_source = 2.5

nfw = mp.SphericalNFW(kappa_s=0.1, scale_radius=1.0)

mass_at_200 = nfw.mass_at_200_for_units(
    unit_mass="solMass",
    unit_length="arcsec",
    redshift_profile=redshift_lens,
    redshift_source=redshift_source,
    cosmology=cosmology,
)

print(mass_at_200)

radius_at_200 = nfw.radius_at_200_for_units(
    unit_mass="solMass",
    unit_length="arcsec",
    redshift_profile=redshift_lens,
    redshift_source=redshift_source,
    cosmology=cosmology,
)

print(radius_at_200)

concentration = nfw.concentration_for_units(
    unit_mass="solMass",
    unit_length="arcsec",
    redshift_profile=redshift_lens,
    redshift_source=redshift_source,
    cosmology=cosmology,
)

print(concentration)

nfw = mp.SphericalTruncatedNFW(kappa_s=0.1, scale_radius=1.0, truncation_radius=1.0)

mass_at_truncation_radius = nfw.mass_at_truncation_radius(
    unit_mass="solMass",
    unit_length="arcsec",
    redshift_profile=redshift_lens,
    redshift_source=redshift_source,
    cosmology=cosmology,
)

print(mass_at_truncation_radius)



from autolens.data.instrument import ccd
from autolens.array import mask as ma
from autolens.lens import ray_tracing, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.lens import lens_data as ld
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import ray_tracing_plotters
from autolens.lens.plotters import lens_fit_plotters

# In this example, we'll fit the ccd data we simulated in the previous exercise. We'll do this using model
# images generated via a tracer, and by comparing to the simulated image we'll get diagostics about the quality of the fit.

# First you need to change the path below to the chapter 1 directory so we can load the data we output previously.
chapter_path = (
    "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_1_introduction/"
)

# The data path specifies where the data was output in the last tutorial, this time in the directory 'chapter_path/data'
data_path = chapter_path + "data/"

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    noise_map_path=data_path + "noise_map.fits",
    psf_path=data_path + "psf.fits",
    pixel_scale=0.1,
)

# 'ccd_data' is a CCDData object, which is a 'package' of all components of the CCD instrument of the lens,
# in particular:

# 1) The image.
# 2) The Point Spread Function (PSF).
# 3) Its noise-map.

print("Image:")
print(ccd_data.image)
print("Noise-Map:")
print(ccd_data.noise_map)
print("PSF:")
print(ccd_data.psf)

# To fit an image, we first specify a mask. A mask describes the sections of the image that we fit.

# Typically, we want to mask regions of the image where the lens and source galaxies are not visible, for example
# at the edges where the signal is entirely background sky and noise.

# For the image we simulated, a 3" circular mask will do the job.

mask = ma.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
)

print(mask)  # 1 = True, which means the pixel is masked. Edge pixels are indeed masked.
print(mask[48:53, 48:53])  # Whereas central pixels are False and therefore unmasked.

# We can use a ccd_plotter to compare the mask and the image - this is useful if we really want to 'tailor' a
# mask to the lensed source's light (which in this example, we won't).
ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)

# We can also use the mask to 'zoom' our plot around the masked region only - meaning that if our image is very large,
# we can focus-in on the lens and source galaxies.

# You'll see this is an option for pretty much every plotter in PyAutoLens, and is something we'll do often throughout
# the tutorials.
ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask, zoom_around_mask=True)

# We can also remove all pixels outside the mask in the plot, meaning bright pixels outside the mask won't impact the
# plot's color range. Again, we'll do this throughout the code.
ccd_plotters.plot_image(
    ccd_data=ccd_data, mask=mask, extract_array_from_mask=True, zoom_around_mask=True
)

# Now we've loaded the ccd data and created a mask, we'll create a 'lens data' object, using the 'lens_data' module.

# A lens data object is a 'package' of all parts of a data-set we need in order to fit it with a lens model:

# 1) The ccd-data, including the image, PSF (so that when we compare a tracer's image to the image instrument we
#    can include blurring due to the telescope optics) and noise-map (so our goodness-of-fit measure accounts for
#    noise in the observations).

# 2) The mask, so that only the regions of the image with a signal are fitted.

# 3) A grid aligned to the ccd-imaging data's pixels, so the tracer's image is generated on the same (masked) grid as
# the image.

lens_data = ld.LensData(ccd_data=ccd_data, mask=mask)

ccd_plotters.plot_image(ccd_data=lens_data.ccd_data)

# By printing its attribute, we can see that it does indeed contain the unmasked image, unmasked noise-map mask,
# psf and so on.
print("Image:")
print(lens_data.unmasked_image)
print()
print("Noise-Map:")
print(lens_data.unmasked_noise_map)
print()
print("PSF:")
print(lens_data.psf)
print()
print("Mask")
print(lens_data.mask_2d)
print()

# The lens_data also contains a masked image, returned below in 2D and 1D.

# On the 2D array, all edge values are masked and are therefore zeros. To see the image values, try changing the
# indexes of the array that are print to the central pixels (e.g. [49, 50])

print("The 2D Masked Image and 1D Image of unmasked entries")
print(lens_data.image(return_in_2d=True).shape)
print(lens_data.image(return_in_2d=True))
print(lens_data.image(return_in_2d=False).shape)
print(lens_data.image(return_in_2d=False))

# There is a masked noise-map, again returned in 2D and 1D with edge values set to zeros.

print("The 2D Masked Noise-Map and 1D Noise-Map of unmasked entries")
print(lens_data.noise_map(return_in_2d=True).shape)
print(lens_data.noise_map(return_in_2d=True))
print(lens_data.noise_map(return_in_2d=False).shape)
print(lens_data.noise_map(return_in_2d=False))

# The lens-data also has a grid, where only coordinates which are not masked are included.
print("Masked Grid")
print(lens_data.grid)

# To fit an image, create an image using a tracer. Lets use the same tracer we simulated the ccd instrument with (thus,
# our fit is 'perfect').

# Its worth noting that below, we use the lens_data's grid to setup the tracer. This ensures that our image-plane
# image is the same resolution and alignment as our lens data's masked image.

lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

source_galaxy = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = ray_tracing.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

ray_tracing_plotters.plot_profile_image(tracer=tracer, grid=lens_data.grid)

# To fit the image, we pass the lens data and tracer to the 'lens_fit' module. This performs the following:

# 1) Blurs the tracer's image with the lens data's PSF, ensuring the telescope optics are included in the fit. This
#    creates the fit's 'model_image'.

# 2) Computes the difference between this model_image and the observed image-data, creating the fit's 'residual_map'.

# 3) Divides the residual-map by the noise-map and squares each value, creating the fit's 'chi_squared_map'.

# 4) Sums up these chi-squared values and converts them to a 'likelihood', which quantifies how good the tracer's fit
#    to the data was (higher likelihood = better fit).

fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# We can print the fit's attributes. As usual, we can choose whether to return the fits in 2d or 1d, and in 2d if we
# don't specify where we'll get all zeros, as the edges were masked:
print("Model-Image:")
print(fit.model_image(return_in_2d=True))
print(fit.model_image(return_in_2d=False))
print()
print("Residual Maps:")
print(fit.residual_map(return_in_2d=True))
print(fit.residual_map(return_in_2d=False))
print()
print("Chi-Squareds Maps:")
print(fit.chi_squared_map(return_in_2d=True))
print(fit.chi_squared_map(return_in_2d=False))

# Of course, the central unmasked pixels have non-zero values.
print("Model-Image Central Pixels:")
model_image = fit.model_image(return_in_2d=True)
print(model_image[48:53, 48:53])
print()

residual_map = fit.residual_map(return_in_2d=True)
print("Residuals Central Pixels:")
print(residual_map[48:53, 48:53])
print()

print("Chi-Squareds Central Pixels:")
chi_squared_map = fit.chi_squared_map(return_in_2d=True)
print(chi_squared_map[48:53, 48:53])

# The fit also gives a likelihood, which is a single-figure estimate of how good the model image fitted the
# simulated image (in unmasked pixels only!).
print("Likelihood:")
print(fit.likelihood)

# We used the same tracer to create and fit the image, giving an excellent fit. The residual-map and chi-squared-map,
# show no signs of the source galaxy's light present, indicating a good fit. This solution will translate to one of
# the highest-likelihood solutions possible.

# Lets change the tracer, so that it's near the correct solution, but slightly off. Below, we slightly offset the
# lens galaxy, by 0.005"

lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.005, 0.005), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

source_galaxy = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tracer = ray_tracing.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# Residuals now appear at the locations of the source galaxy, increasing the chi-squared values (which determine our
# likelihood).

# Lets compare the likelihood to the value we computed above (which was 11697.24):
print("Previous Likelihood:")
print(11697.24)
print("New Likelihood:")
print(fit.likelihood)

# It decreases! As expected, this model us a worse fit to the data.

# Lets change the tracer, one more time, to a solution nowhere near the correct one.
lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.005, 0.005), einstein_radius=1.3, axis_ratio=0.8, phi=45.0
    ),
)

source_galaxy = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.7,
        phi=65.0,
        intensity=1.0,
        effective_radius=0.4,
        sersic_index=3.5,
    ),
)

tracer = ray_tracing.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# Clearly, the model provides a terrible fit and this tracer is not a plausible representation of
# the image-data (of course, we already knew that, given that we simulated it!)

# The likelihood drops dramatically, as expected.
print("Previous Likelihoods:")
print(11697.24)
print(10319.44)
print("New Likelihood:")
print(fit.likelihood)

# Congratulations, you've fitted your first strong lens with PyAutoLens! Perform the following exercises:

# 1) In this example, we 'knew' the correct solution, because we simulated the lens ourselves. In the real Universe,
#    we have no idea what the correct solution is. How would you go about finding the correct solution?
#    Could you find a solution that fits the data reasonable through trial and error?



import autofit as af
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.array import grids
from autolens.lens import ray_tracing
from autolens.data.plotters import ccd_plotters


class CCDObservation(object):
    def __init__(self, shape, pixel_scale, psf, exposure_time, background_sky_level):
        """A class representing a CCD observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        shape : (int, int)
            The shape of the observation. Note that we do not simulate a full CCD frame (e.g. 2000 x 2000 pixels for \
            Hubble imaging), but instead just a cut-out around the strong lens.
        pixel_scale : float
            The size of each pixel in arc seconds.
        psf : PSF
            An array describing the PSF kernel of the image.
        exposure_time : float
            The exposure time of an observation using this instrument.
        background_sky_level : float
            The level of the background sky of an observationg using this instrument.
        """
        self.shape = shape
        self.pixel_scale = pixel_scale
        self.psf = psf
        self.exposure_time = exposure_time
        self.background_sky_level = background_sky_level

    def simulate_ccd_data_from_galaxies(
        self,
        galaxies,
        sub_grid_size=16,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        should_plot_ccd_data=False,
    ):
        """Simulate CCD instrument for this instrument, as follows:

        1)  Setup the image-plane al.ogrid of the CCD array, which defines the coordinates used for the ray-tracing.

        2) Use this grid and the lens and source galaxies to setup a tracer, which generates the image of \
           the simulated CCD instrument.

        3) Simulate the CCD instrument, using a special image which ensures edge-effects don't
           degrade simulation of the telescope optics (e.g. the PSF convolution).

        4) Plot the image using Matplotlib, if the plot_ccd bool is True.

        5) Output the data to .fits format if a data_path and data_name are specified. Otherwise, return the simulated \
           ccd instrument instance."""

        grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=self.shape,
            pixel_scale=self.psf.pixel_scale,
            sub_grid_size=sub_grid_size,
        )

        tracer = ray_tracing.Tracer.from_galaxies(galaxies=galaxies)

        ccd_data = ccd.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
            tracer=tracer,
            grid=grid,
            pixel_scale=self.pixel_scale,
            exposure_time=self.exposure_time,
            psf=self.psf,
            background_sky_level=self.background_sky_level,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

        if should_plot_ccd_data:
            ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

        return ccd_data

    def simulate_ccd_data_from_galaxies_and_write_to_fits(
        self,
        galaxies,
        data_path,
        data_name,
        sub_grid_size=16,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        should_plot_ccd_data=False,
    ):
        """Simulate CCD instrument for this instrument, as follows:

        1)  Setup the image-plane al.ogrid of the CCD array, which defines the coordinates used for the ray-tracing.

        2) Use this grid and the lens and source galaxies to setup a tracer, which generates the image of \
           the simulated CCD instrument.

        3) Simulate the CCD instrument, using a special image which ensures edge-effects don't
           degrade simulation of the telescope optics (e.g. the PSF convolution).

        4) Plot the image using Matplotlib, if the plot_ccd bool is True.

        5) Output the data to .fits format if a data_path and data_name are specified. Otherwise, return the simulated \
           ccd instrument instance."""

        ccd_data = self.simulate_ccd_data_from_galaxies(
            galaxies=galaxies,
            sub_grid_size=sub_grid_size,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
            should_plot_ccd_data=should_plot_ccd_data,
        )

        data_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=data_path, folder_names=[data_name]
        )

        ccd.output_ccd_data_to_fits(
            ccd_data=ccd_data,
            image_path=data_output_path + "image.fits",
            psf_path=data_output_path + "psf.fits",
            noise_map_path=data_output_path + "noise_map.fits",
            exposure_time_map_path=data_output_path + "exposure_time_map.fits",
            background_noise_map_path=data_output_path + "background_noise_map.fits",
            poisson_noise_map_path=data_output_path + "poisson_noise_map.fits",
            background_sky_map_path=data_output_path + "background_sky_map.fits",
            overwrite=True,
        )

    @classmethod
    def lsst(
        cls,
        shape=(101, 101),
        pixel_scale=0.2,
        psf_shape=(31, 31),
        psf_sigma=0.5,
        exposure_time=100.0,
        background_sky_level=1.0,
    ):
        """Default settings for an observation with the Large Synotpic Survey Telescope.

        This can be customized by over-riding the default input values."""
        psf = abstract_data.PSF.from_gaussian(
            shape=psf_shape, sigma=psf_sigma, pixel_scale=pixel_scale
        )
        return CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

    @classmethod
    def euclid(
        cls,
        shape=(151, 151),
        pixel_scale=0.1,
        psf_shape=(31, 31),
        psf_sigma=0.1,
        exposure_time=565.0,
        background_sky_level=1.0,
    ):
        """Default settings for an observation with the Euclid space satellite.

        This can be customized by over-riding the default input values."""
        psf = abstract_data.PSF.from_gaussian(
            shape=psf_shape, sigma=psf_sigma, pixel_scale=pixel_scale
        )
        return CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

    @classmethod
    def hst(
        cls,
        shape=(251, 251),
        pixel_scale=0.05,
        psf_shape=(31, 31),
        psf_sigma=0.05,
        exposure_time=2000.0,
        background_sky_level=1.0,
    ):
        """Default settings for an observation with the Hubble Space Telescope.

        This can be customized by over-riding the default input values."""
        psf = abstract_data.PSF.from_gaussian(
            shape=psf_shape, sigma=psf_sigma, pixel_scale=pixel_scale
        )
        return CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

    @classmethod
    def hst_up_sampled(
        cls,
        shape=(401, 401),
        pixel_scale=0.03,
        psf_shape=(31, 31),
        psf_sigma=0.05,
        exposure_time=2000.0,
        background_sky_level=1.0,
    ):
        """Default settings for an observation with the Hubble Space Telescope which has been upscaled to a higher \
        pixel-scale to better sample the PSF.

        This can be customized by over-riding the default input values."""
        psf = abstract_data.PSF.from_gaussian(
            shape=psf_shape, sigma=psf_sigma, pixel_scale=pixel_scale
        )
        return CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )

    @classmethod
    def keck_adaptive_optics(
        cls,
        shape=(751, 751),
        pixel_scale=0.01,
        psf_shape=(31, 31),
        psf_sigma=0.025,
        exposure_time=1000.0,
        background_sky_level=1.0,
    ):
        """Default settings for an observation using Keck Adaptive Optics imaging.

        This can be customized by over-riding the default input values."""
        psf = abstract_data.PSF.from_gaussian(
            shape=psf_shape, sigma=psf_sigma, pixel_scale=pixel_scale
        )
        return CCDObservation(
            shape=shape,
            pixel_scale=pixel_scale,
            psf=psf,
            exposure_time=exposure_time,
            background_sky_level=background_sky_level,
        )



import autofit as af
import autolens as al
from workspace.tools.loading_and_preparing_data import simulate_data

# To model data with PyAutoLens you first need to ensure it is in a format suitable for lens modeling. This tutorial
# takes you through data preparation, introducing PyAutoLens's built in tools that convert data to a suitable format.

# First, lets setup the path to our current working directory. I recommend you use the 'autolens_workspace' directory
# and place data in the 'autolens_workspace/data' directory.

# for this tutorial, we'll use the 'autolens_workspace/tools/loading_and_preparing_data/data' directory. The folder
# 'data' contains example data we'll use in this tutorial.

path = (
    "path/to/AutoLens/autolens_workspace/tools/loading_and_preparing_data/"
)  # <----- You must include this slash on the end
path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/loading_and_preparing_data/"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=path, folder_names=["data"]
)

# This populates the 'data' path with example simulated ccd data-sets.

simulate_data.simulate_all_ccd_data(data_path=data_path)

# First, lets load a data-set using the 'load_ccd_data_from_fits' function of the ccd module. This
# data-set represents a good data-reduction that conforms to the formatting standards I describe in this tutorial!
ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data"]
)

ccd_data = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    pixel_scale=0.1,
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# If your data comes in one .fits file spread across multiple hdus you can specify the hdus of each image instead.
ccd_data = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "multiple_hdus.fits",
    image_hdu=0,
    noise_map_path=ccd_data_path + "multiple_hdus.fits",
    noise_map_hdu=1,
    psf_path=ccd_data_path + "multiple_hdus.fits",
    psf_hdu=2,
    pixel_scale=0.1,
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Lets think about the format of our data. There are numerous reasons why the image we just looked at is a good data-set
# for lens modeling. I strongly recommend you reduce your data to conform to the standards discussed below - it'll make
# your time using PyAutoLens a lot simpler.

# However, you may not have access to the data-reduction tools that made the data, so we've included a number of
# in-built functions in PyAutoLens to convert the data to a good format for you. However, your life will be much easier
# if you can just reduce it this way in the first place!

# 1) Brightness units - the image's flux and noise-map values are in units of electrons per second (not electrons,
#    counts, ADU's etc.). Although PyAutoLens can technically perform an analysis using other units, the default
#    settings assume the image is in electrons per second (e.g. the priors on light profile image and
#    regularization coefficient). Thus, images not in electrons per second should be converted!

# Lets look at an image that is in units of counts - its easy to tell because the peak values are in the 1000's or
# 10000's.
ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data_in_counts"]
)

ccd_data_in_counts = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_in_counts)

# If your data is in counts you can convert it to electrons per second by supplying the function above with an
# exposure time and using the 'convert_arrays_from_counts' boolean flag.
ccd_data_converted_to_eps = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    exposure_time_map_from_single_value=1000.0,
    convert_from_electrons=True,
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_converted_to_eps)

# The effective exposure time in each pixel may vary. This occurs when data is reduced using 'dithering' and 'drizzling'.
# If you have access to an effective exposure-time map, you can use this to convert the image to electrons per second
# instead.
ccd_data_converted_to_eps = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    exposure_time_map_path=ccd_data_path + "exposure_time_map.fits",
    convert_from_electrons=True,
)
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_converted_to_eps)

# 2) Postage stamp size - The bigger the postage stamp cut-out of the image the more memory it requires to store it.
#    Why keep the edges surrounding the lens if they are masked out anyway?

#    Lets look at an example of a very large postage stamp - we can barely even see the lens and source galaxies!

ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data_with_large_stamp"]
)

ccd_data_large_stamp = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_large_stamp)

#  If you have a large postage stamp you can trim it when you load the data by specifying a new image size in pixels.
#  This will also trim the noise-map, exposoure time map and other arrays which are the same dimensions as the image.
#  This trimming is centred on the image.
ccd_data_large_stamp_trimmed = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    resized_ccd_shape=(101, 101),
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_large_stamp_trimmed)

# 3) Postage stamp size (again). The stamp may also be too small - for example it must have enough padding in the border
#    that our mask includes all pixels with signal. In fact, this padding must also include the a 'blurring region',
#    corresponding to all unmasked image pixels where light blurs into the masks after PSF convolution. Thus, we may
#    need to pad an image to include this region.

ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data_with_small_stamp"]
)

# This image is an example of a stamp which is big enough to contain the lens and source galaxies, but when we
# apply a sensible masks we get an error, because the masks's blurring region goes into the edge of the image.
ccd_data_small_stamp = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_small_stamp)

# If we apply a masks to this image we get an error when we try to use it to set up a lensing image because its
# blurring region hits the image edge.

mask = al.Mask.circular(
    shape=ccd_data_small_stamp.shape,
    pixel_scale=ccd_data_small_stamp.pixel_scale,
    radius_arcsec=2.0,
)

# This gives an error because the mask's blurring region hits an edge.
# lens_data = ld.LensData(ccd_data=ccd_data_small_stamp, mask=mask)

# We overcome this using the same input as before. However, now, the resized image shape is bigger than the image,
# thus a padding of zeros is introduced to the edges.
ccd_data_small_stamp_padded = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    resized_ccd_shape=(140, 140),
)

mask = al.Mask.circular(
    shape=ccd_data_small_stamp_padded.shape,
    pixel_scale=ccd_data_small_stamp_padded.pixel_scale,
    radius_arcsec=2.0,
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_small_stamp_padded, mask=mask)

# This no longer gives an error!
lens_data = al.LensData(ccd_data=ccd_data_small_stamp_padded, mask=mask)

########## IVE INCLUDED THE TEXT FOR 5 BELOW SO YOU CAN BE AWARE OF CENTERING, BUT THE BUILT IN FUNCTIONALITY FOR #####
########## RECENTERING CURRENTLY DOES NOT WORK :( ###########

# 5) Lens Galaxy Centering - The lens galaxy should be in the centre of the image as opposed to a corner. This ensures
#    the origin of the lens galaxy's light and mass profiles are near the origin (0.0", 0.0") of the grid used to perform
#    ray-tracing. The defaults priors on light and mass profiles assume a origin of (0.0", 0.0").

# Lets look at an off-center image - clearly both the lens galaxy and Einstein ring are offset in the positive y and x d
# directions.

# ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(path=data_path,
#                                                                           folder_names=['ccd_data_offset_centre'])

# ccd_data_offset_centre = al.load_ccd_data_from_fits(image_path=path+'image.fits', pixel_scale=0.1,
#                                   noise_map_path=path+'noise_map.fits',
#                                   psf_path=path+'psf.fits')
# ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_offset_centre)

# We can address this by using supplying a new centre for the image, in pixels. We also supply the resized shape, to
# instruct the code whether it should trim the image or pad the edges that now arise due to recentering.

# ccd_data_recentred_pixels = al.load_ccd_data_from_fits(image_path=path+'image.fits', pixel_scale=0.1,
#                                             noise_map_path=path+'noise_map.fits',
#                                             psf_path=path+'psf.fits',
#                                             resized_ccd_shape=(100, 100),
#                                             resized_ccd_centre_pixels=(0, 0))
# #                                            resized_ccd_centre_arc_seconds=(1.0, 1.0))
# print(ccd_data_recentred_pixels.shape)
# ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_recentred_pixels)

# 6) The noise-map values are the RMS standard deviation in every pixel (and not the variances, HST WHT-map values,
#    etc.). You MUST be 100% certain that the noise_map map is the RMS standard deviations or else your analysis will
#    be incorrect.

# There are many different ways the noise-map can be reduced. We are aiming to include conversion functions for all
# common data-reductions. Currently, we have a function to convert an image from a HST WHT map, where
# RMS SD = 1.0/ sqrt(WHT). This can be called using the 'convert_noise_map_from_weight_map' flag.

ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data_with_large_stamp"]
)

ccd_data_noise_from_wht = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    convert_noise_map_from_weight_map=True,
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_noise_from_wht)

# (I don't currently have an example image in WHT for this tutorial, but the function above will work. Above, it
# actually converts an accurate noise-map to an inverse WHT map!

# 7) The PSF zooms around its central core, which is the most important region for strong lens modeling. By
#    default, the size of the PSF image is used to perform convolution. The larger this stamp, the longer this
#    convolution will take to run. In general, we would recommend the PSF size is 21 x 21.

#    Lets look at an image where a large PSF kernel is loaded.

ccd_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_path, folder_names=["ccd_data_with_large_psf"]
)

ccd_data_with_large_psf = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_with_large_psf)

# We can resize a psf the same way that we resize an image.
ccd_data_with_trimmed_psf = al.load_ccd_data_from_fits(
    image_path=ccd_data_path + "image.fits",
    pixel_scale=0.1,
    noise_map_path=ccd_data_path + "noise_map.fits",
    psf_path=ccd_data_path + "psf.fits",
    resized_psf_shape=(21, 21),
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data_with_trimmed_psf)

# 8) The PSF dimensions are odd x odd (21 x 21). It is important that the PSF dimensions are odd, because even-sized
#    PSF kernels introduce a half-pixel offset in the convolution routine which can lead to systematics in the lens
#    analysis.

# We do not currently have built-in functionality to address this issue. Therefore, if your PSF has an even
# dimension you must manually trim and recentre it. If you need help on doing this, contact me on the PyAutoLens
# SLACK channel, as I'll have already written the routine to do this by the time you read this tutorial!



import autofit as af
import autolens as al


# In this pipeline, we'll demonstrate signal-to-noise limiting - which allows us to fit ccd instrument where the noise-map
# has been increased to cap the maximum signal-to-noise value in the image in a phase of the pipeline. In this example,
# we will perform an initial analysis on an image with a signal-to-noise limit of 20.0, and then fit the image using
# the unscaled signal-to-noise map.

# Why would you want to limit the signal to noise? There are two reasons :
#
# 1) The model fitting, may be subject to  over-fitting the highest signal-to-noise regions of the image instead of
#    providing a global fit to the entire image. For example, if a lensed source has 4 really bright, compact, high
#    S/N images which are not fitted perfectly by the model, their high chi-squared contribution will drive the model
#    fit to place more light in those regions, ignoring the lensed source's lower S/N more extended arcs. Limiting the
#    S/N of these high S/N regions will reduce over-fitting. This is also important lens light subtractions which are
#    not perfect, which leave large chi-squared residuals due to the lens light's high S/N.

#    To learn more about this over-fitting problem, checkout chapter 5 of the 'HowToLens' lecture series.

# 2) If the model-fit has extremely large chi-squared values due to the high S/N of the data, this means the non-linear
#    search will take a long time mapping_util out this 'extreme' parameter space. In the early phases of a pipeline this
#    often isn't necessary, therefore a signal-to-noise limit can reduce the time an analysis takes to converge.

# Whilst signal to noise limits can be manually specified in the pipeline, in this example we will make the signal to
# noise limit an input parameter of the pipeline. This means we can run the pipeline with different signal to noise
# limits for different runners.

# We will also use phase tagging to ensure phases which use a signal to noise limit have a tag in their path, so it is
# clear what settings a phases has when it uses this feature.

# We'll perform a basic analysis which fits a lensed source galaxy using a parametric light profile where
# the lens's light is omitted. This pipeline uses two phases:

# Phase 1:

# Description: initialize the lens mass model and source light profile using x1 source with a signal to noise limit of 10.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses a signal to noise limit of 10

# Phase 2:

# Description: Fits the lens and source model using the true signal to noise map
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: No signal to noise limit.


def make_pipeline(phase_folders=None, signal_to_noise_limit=20.0):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_feature__signal_to_noise_limit"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # When a phase is passed a signal_to_noise_limit, a settings tag is automatically generated and added to the phase
    # path,to make it clear what signal-to-noise limit was used. The settings tag, phase name and phase paths are shown
    # for 3 example signal-to-noise limits:

    # signal_to_noise_limit=2 -> phase_path=phase_name/settings_snr_2
    # signal_to_noise_limit=3 -> phase_path=phase_name/settings_snr_3

    # If the signal_to_noise_limit is None, the tag is an empty string, thus not changing the settings tag:

    # signal_to_noise_limit=None -> phase_path=phase_name/settings

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/pipeline_name/pipeline_tag/phase_name/phase_tag//'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return al.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Use a signal-to-noise limit of 20.0

    class LensSourceX1Phase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens.mass.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.lens.mass.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourceX1Phase(
        phase_name="phase_1__x1_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        signal_to_noise_limit=signal_to_noise_limit,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Omit the signal-to-noise limit, thus performing the modeling at the image's native signal-to-noise.

    phase2 = al.PhaseImaging(
        phase_name="phase_2__x2_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=phase1.result.variable.galaxies.lens.mass,
                shear=phase1.result.variable.galaxies.lens.shear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=phase1.result.variable.galaxies.source.light
            ),
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return al.PipelineImaging(pipeline_name, phase1, phase2)



import numpy as np
from skimage.transform import rescale

from autolens import decorator_util
from autolens import exc
from autolens.array.util import grid_util
from autolens.array.mapping_util import mask_mapping_util


@decorator_util.jit()
def mask_centres_from_shape_pixel_scale_and_centre(shape, pixel_scale, centre):
    """Determine the (y,x) arc-second central coordinates of a mask from its shape, pixel-scales and centre.

     The coordinate system is defined such that the positive y axis is up and positive x axis is right.

    Parameters
     ----------
    shape : (int, int)
        The (y,x) shape of the 2D array the arc-second centre is computed for.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D array.
    centre : (float, flloat)
        The (y,x) centre of the 2D mask.

    Returns
    --------
    tuple (float, float)
        The (y,x) arc-second central coordinates of the input array.

    Examples
    --------
    centres_arcsec = centres_from_shape_pixel_scales_and_centre(shape=(5,5), pixel_scales=(0.5, 0.5), centre=(0.0, 0.0))
    """
    y_centre_arcsec = (float(shape[0] - 1) / 2) - (centre[0] / pixel_scale)
    x_centre_arcsec = (float(shape[1] - 1) / 2) + (centre[1] / pixel_scale)

    return (y_centre_arcsec, x_centre_arcsec)


@decorator_util.jit()
def total_pixels_from_mask(mask):
    """Compute the total number of unmasked pixels in a mask.

    Parameters
     ----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked and included when counting pixels.

    Returns
    --------
    int
        The total number of pixels that are unmasked.

    Examples
    --------

    mask = np.array([[True, False, True],
                 [False, False, False]
                 [True, False, True]])

    total_regular_pixels = total_regular_pixels_from_mask(mask=mask)
    """

    total_regular_pixels = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                total_regular_pixels += 1

    return total_regular_pixels


@decorator_util.jit()
def total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size):
    """Compute the total number of sub-pixels in unmasked pixels in a mask.
    
    Parameters
     ----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked and included when counting sub pixels.
    sub_grid_size : int
        The size of the sub-grid that each pixel of the 2D mask array is divided into.

    Returns
    --------
    int
        The total number of sub pixels that are unmasked.

    Examples
    --------

    mask = np.array([[True, False, True],
                     [False, False, False]
                     [True, False, True]])

    total_sub_pixels = total_sub_pixels_from_mask(mask=mask, sub_grid_size=2)
    """
    return total_pixels_from_mask(mask) * sub_grid_size ** 2


@decorator_util.jit()
def total_sparse_pixels_from_mask(mask, unmasked_sparse_grid_pixel_centres):
    """Given the full (i.e. without removing pixels which are outside the mask) pixelization grid's pixel \ 
    center and the mask, compute the total number of pixels which are within the mask and thus used \ \
    by the pixelization grid.

    Parameters
    -----------
    mask : ccd.mask.Mask
        The mask within which pixelization pixels must be inside
    unmasked_sparse_grid_pixel_centres : ndarray
        The centres of the unmasked pixelization grid pixels.
    """

    total_sparse_pixels = 0

    for unmasked_sparse_pixel_index in range(
        unmasked_sparse_grid_pixel_centres.shape[0]
    ):

        y = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 0]
        x = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 1]

        if not mask[y, x]:
            total_sparse_pixels += 1

    return total_sparse_pixels


@decorator_util.jit()
def mask_circular_from_shape_pixel_scale_and_radius(
    shape, pixel_scale, radius_arcsec, centre=(0.0, 0.0)
):
    """Compute a circular mask from the 2D mask array shape and radius of the circle.

    This creates a 2D array where all values within the mask radius are unmasked and therefore *False*.

    Parameters
     ----------
    shape: (int, int)
        The (y,x) shape of the mask in units of pixels.
    pixel_scale: float
        The arc-second to pixel conversion factor of each pixel.
    radius_arcsec : float
        The radius (in arc seconds) of the circle within which pixels unmasked.
    centre: (float, float)
        The centre of the circle used to mask pixels.

    Returns
    --------
    ndarray
        The 2D mask array whose central pixels are masked as a circle.

    Examples
    --------
    mask = mask_circular_from_shape_pixel_scale_and_radius( \
        shape=(10, 10), pixel_scale=0.1, radius_arcsec=0.5, centre=(0.0, 0.0))
    """

    mask = np.full(shape, True)

    centres_arcsec = mask_centres_from_shape_pixel_scale_and_centre(
        shape=mask.shape, pixel_scale=pixel_scale, centre=centre
    )

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - centres_arcsec[0]) * pixel_scale
            x_arcsec = (x - centres_arcsec[1]) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if r_arcsec <= radius_arcsec:
                mask[y, x] = False

    return mask


@decorator_util.jit()
def mask_circular_annular_from_shape_pixel_scale_and_radii(
    shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec, centre=(0.0, 0.0)
):
    """Compute an annular mask from an input inner and outer mask radius and shape.

    This creates a 2D array where all values within the inner and outer radii are unmasked and therefore *False*.

    Parameters
     ----------
    shape : (int, int)
        The (y,x) shape of the mask in units of pixels.
    pixel_scale: float
        The arc-second to pixel conversion factor of each pixel.
    inner_radius_arcsec : float
        The radius (in arc seconds) of the inner circle outside of which pixels are unmasked.
    outer_radius_arcsec : float
        The radius (in arc seconds) of the outer circle within which pixels are unmasked.
    centre: (float, float)
        The centre of the annulus used to mask pixels.

    Returns
    --------
    ndarray
        The 2D mask array whose central pixels are masked as a annulus.

    Examples
    --------
    mask = mask_annnular_from_shape_pixel_scale_and_radius( \
        shape=(10, 10), pixel_scale=0.1, inner_radius_arcsec=0.5, outer_radius_arcsec=1.5, centre=(0.0, 0.0))
    """

    mask = np.full(shape, True)

    centres_arcsec = mask_centres_from_shape_pixel_scale_and_centre(
        shape=mask.shape, pixel_scale=pixel_scale, centre=centre
    )

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - centres_arcsec[0]) * pixel_scale
            x_arcsec = (x - centres_arcsec[1]) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if outer_radius_arcsec >= r_arcsec >= inner_radius_arcsec:
                mask[y, x] = False

    return mask


@decorator_util.jit()
def mask_circular_anti_annular_from_shape_pixel_scale_and_radii(
    shape,
    pixel_scale,
    inner_radius_arcsec,
    outer_radius_arcsec,
    outer_radius_2_arcsec,
    centre=(0.0, 0.0),
):
    """Compute an annular mask from an input inner and outer mask radius and shape."""

    mask = np.full(shape, True)

    centres_arcsec = mask_centres_from_shape_pixel_scale_and_centre(
        shape=mask.shape, pixel_scale=pixel_scale, centre=centre
    )

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - centres_arcsec[0]) * pixel_scale
            x_arcsec = (x - centres_arcsec[1]) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if (
                inner_radius_arcsec >= r_arcsec
                or outer_radius_2_arcsec >= r_arcsec >= outer_radius_arcsec
            ):
                mask[y, x] = False

    return mask


@decorator_util.jit()
def elliptical_radius_from_y_x_phi_and_axis_ratio(y_arcsec, x_arcsec, phi, axis_ratio):
    r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

    theta_rotated = np.arctan2(y_arcsec, x_arcsec) + np.radians(phi)

    y_arcsec_elliptical = r_arcsec * np.sin(theta_rotated)
    x_arcsec_elliptical = r_arcsec * np.cos(theta_rotated)

    return np.sqrt(
        x_arcsec_elliptical ** 2.0 + (y_arcsec_elliptical / axis_ratio) ** 2.0
    )


@decorator_util.jit()
def mask_elliptical_from_shape_pixel_scale_and_radius(
    shape, pixel_scale, major_axis_radius_arcsec, axis_ratio, phi, centre=(0.0, 0.0)
):
    """Compute an elliptical mask from an input major-axis mask radius, axis-ratio, rotational angle phi, shape and \
    centre.

    This creates a 2D array where all values within the ellipse are unmasked and therefore *False*.

    Parameters
     ----------
    shape: (int, int)
        The (y,x) shape of the mask in units of pixels.
    pixel_scale: float
        The arc-second to pixel conversion factor of each pixel.
    major_axis_radius_arcsec : float
        The major-axis (in arc seconds) of the ellipse within which pixels are unmasked.
    axis_ratio : float
        The axis-ratio of the ellipse within which pixels are unmasked.
    phi : float
        The rotation angle of the ellipse within which pixels are unmasked, (counter-clockwise from the positive \
         x-axis).
    centre: (float, float)
        The centre of the ellipse used to mask pixels.

    Returns
    --------
    ndarray
        The 2D mask array whose central pixels are masked as an ellipse.

    Examples
    --------
    mask = mask_elliptical_from_shape_pixel_scale_and_radius( \
        shape=(10, 10), pixel_scale=0.1, major_axis_radius_arcsec=0.5, axis_ratio=0.5, phi=45.0, centre=(0.0, 0.0))
    """

    mask = np.full(shape, True)

    centres_arcsec = mask_centres_from_shape_pixel_scale_and_centre(
        shape=mask.shape, pixel_scale=pixel_scale, centre=centre
    )

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - centres_arcsec[0]) * pixel_scale
            x_arcsec = (x - centres_arcsec[1]) * pixel_scale

            r_arcsec_elliptical = elliptical_radius_from_y_x_phi_and_axis_ratio(
                y_arcsec, x_arcsec, phi, axis_ratio
            )

            if r_arcsec_elliptical <= major_axis_radius_arcsec:
                mask[y, x] = False

    return mask


@decorator_util.jit()
def mask_elliptical_annular_from_shape_pixel_scale_and_radius(
    shape,
    pixel_scale,
    inner_major_axis_radius_arcsec,
    inner_axis_ratio,
    inner_phi,
    outer_major_axis_radius_arcsec,
    outer_axis_ratio,
    outer_phi,
    centre=(0.0, 0.0),
):
    """Compute an elliptical annular mask from an input major-axis mask radius, axis-ratio, rotational angle phi for \
     both the inner and outer elliptical annuli and a shape and centre for the mask.

    This creates a 2D array where all values within the elliptical annuli are unmasked and therefore *False*.

    Parameters
     ----------
    shape: (int, int)
        The (y,x) shape of the mask in units of pixels.
    pixel_scale: float
        The arc-second to pixel conversion factor of each pixel.
    inner_major_axis_radius_arcsec : float
        The major-axis (in arc seconds) of the inner ellipse within which pixels are masked.
    inner_axis_ratio : float
        The axis-ratio of the inner ellipse within which pixels are masked.
    inner_phi : float
        The rotation angle of the inner ellipse within which pixels are masked, (counter-clockwise from the \
        positive x-axis).
    outer_major_axis_radius_arcsec : float
        The major-axis (in arc seconds) of the outer ellipse within which pixels are unmasked.
    outer_axis_ratio : float
        The axis-ratio of the outer ellipse within which pixels are unmasked.
    outer_phi : float
        The rotation angle of the outer ellipse within which pixels are unmasked, (counter-clockwise from the \
        positive x-axis).
    centre: (float, float)
        The centre of the elliptical annuli used to mask pixels.

    Returns
    --------
    ndarray
        The 2D mask array whose elliptical annuli pixels are masked

    Examples
    --------
    mask = mask_elliptical_annuli_from_shape_pixel_scale_and_radius( \
        shape=(10, 10), pixel_scale=0.1,
         inner_major_axis_radius_arcsec=0.5, inner_axis_ratio=0.5, inner_phi=45.0,
         outer_major_axis_radius_arcsec=1.5, outer_axis_ratio=0.8, outer_phi=90.0,
         centre=(0.0, 0.0))
    """

    mask = np.full(shape, True)

    centres_arcsec = mask_centres_from_shape_pixel_scale_and_centre(
        shape=mask.shape, pixel_scale=pixel_scale, centre=centre
    )

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - centres_arcsec[0]) * pixel_scale
            x_arcsec = (x - centres_arcsec[1]) * pixel_scale

            inner_r_arcsec_elliptical = elliptical_radius_from_y_x_phi_and_axis_ratio(
                y_arcsec, x_arcsec, inner_phi, inner_axis_ratio
            )

            outer_r_arcsec_elliptical = elliptical_radius_from_y_x_phi_and_axis_ratio(
                y_arcsec, x_arcsec, outer_phi, outer_axis_ratio
            )

            if (
                inner_r_arcsec_elliptical >= inner_major_axis_radius_arcsec
                and outer_r_arcsec_elliptical <= outer_major_axis_radius_arcsec
            ):
                mask[y, x] = False

    return mask


@decorator_util.jit()
def blurring_mask_from_mask_and_psf_shape(mask, psf_shape):
    """Compute a blurring mask from an input mask and psf shape.

    The blurring mask corresponds to all pixels which are outside of the mask but will have a fraction of their \
    light blur into the masked region due to PSF convolution. The PSF shape is used to determine which pixels these are.
    
    If a pixel is identified which is outside the 2D dimensionos of the input mask, an error is raised and the user \
    should pad the input mask (and associated images).
    
    Parameters
    -----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked.
    psf_shape : (int, int)
        The 2D shape of the PSF which is used to compute the blurring mask.
        
    Returns
    --------
    ndarray
        The 2D blurring mask array whose unmasked values (*False*) correspond to where the mask will have PSF light \
        blurred into them.

    Examples
    --------
    mask = np.array([[True, True, True],
                     [True, False, True]
                     [True, True, True]])      
    
    blurring_mask = blurring_mask_from_mask_and_psf_shape(mask=mask, psf_shape=(3,3)) 
    
    """

    blurring_mask = np.full(mask.shape, True)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range((-psf_shape[0] + 1) // 2, (psf_shape[0] + 1) // 2):
                    for x1 in range((-psf_shape[1] + 1) // 2, (psf_shape[1] + 1) // 2):
                        if (
                            0 <= x + x1 <= mask.shape[1] - 1
                            and 0 <= y + y1 <= mask.shape[0] - 1
                        ):
                            if mask[y + y1, x + x1]:
                                blurring_mask[y + y1, x + x1] = False
                        else:
                            raise exc.MaskException(
                                "setup_blurring_mask extends beyond the sub_grid_size "
                                "of the mask - pad the datas array before masking"
                            )

    return blurring_mask


@decorator_util.jit()
def mask_from_shape_and_mask_1d_index_to_mask_2d_index(
    shape, mask_1d_index_to_mask_2d_index
):
    """For a 1D array that was computed by mapping_util unmasked values from a 2D array of shape (rows, columns), map its \
    indexes back to the original 2D array to create the origianl 2D mask.

    This uses a 1D array 'one_to_two' where each index gives the 2D pixel indexes of the 1D array's unmasked pixels, \
    for example:

    - If one_to_two[0] = [0,0], the first value of the 1D array maps to the pixel [0,0] of the 2D array.
    - If one_to_two[1] = [0,1], the second value of the 1D array maps to the pixel [0,1] of the 2D array.
    - If one_to_two[4] = [1,1], the fifth value of the 1D array maps to the pixel [1,1] of the 2D array.

    Parameters
     ----------
    shape : (int, int)
        The shape of the 2D array which the pixels are defined on.
    mask_1d_index_to_mask_2d_index : ndarray
        An array describing the 2D array index that every 1D array index maps too.

    Returns
    --------
    ndarray
        A 2D mask array where unmasked values are *False*.

    Examples
    --------
    one_to_two = np.array([[0,1], [1,0], [1,1], [1,2], [2,1]])

    mask = mask_from_shape_and_one_to_two(shape=(3,3), one_to_two=one_to_two)
    """

    mask = np.ones(shape)

    for index in range(len(mask_1d_index_to_mask_2d_index)):
        mask[
            mask_1d_index_to_mask_2d_index[index, 0],
            mask_1d_index_to_mask_2d_index[index, 1],
        ] = False

    return mask


@decorator_util.jit()
def total_edge_pixels_from_mask(mask):
    """Compute the total number of borders-pixels in a mask."""

    edge_pixel_total = 0

    for y in range(1, mask.shape[0] - 1):
        for x in range(1, mask.shape[1] - 1):
            if not mask[y, x]:
                if (
                    mask[y + 1, x]
                    or mask[y - 1, x]
                    or mask[y, x + 1]
                    or mask[y, x - 1]
                    or mask[y + 1, x + 1]
                    or mask[y + 1, x - 1]
                    or mask[y - 1, x + 1]
                    or mask[y - 1, x - 1]
                ):
                    edge_pixel_total += 1

    return edge_pixel_total


@decorator_util.jit()
def edge_1d_indexes_from_mask(mask):
    """Compute a 1D array listing all edge pixel indexes in the mask. An edge pixel is a pixel which is not fully \
    surrounding by False mask values i.e. it is on an edge."""

    edge_pixel_total = total_edge_pixels_from_mask(mask)

    edge_pixels = np.zeros(edge_pixel_total)
    edge_index = 0
    regular_index = 0

    for y in range(1, mask.shape[0] - 1):
        for x in range(1, mask.shape[1] - 1):
            if not mask[y, x]:
                if (
                    mask[y + 1, x]
                    or mask[y - 1, x]
                    or mask[y, x + 1]
                    or mask[y, x - 1]
                    or mask[y + 1, x + 1]
                    or mask[y + 1, x - 1]
                    or mask[y - 1, x + 1]
                    or mask[y - 1, x - 1]
                ):
                    edge_pixels[edge_index] = regular_index
                    edge_index += 1

                regular_index += 1

    return edge_pixels


@decorator_util.jit()
def check_if_border_pixel(mask, edge_pixel_1d, mask_1d_index_to_mask_2d_index):
    edge_pixel_index = int(edge_pixel_1d)

    y = int(mask_1d_index_to_mask_2d_index[edge_pixel_index, 0])
    x = int(mask_1d_index_to_mask_2d_index[edge_pixel_index, 1])

    if (
        np.sum(mask[0:y, x]) == y
        or np.sum(mask[y, x : mask.shape[1]]) == mask.shape[1] - x - 1
        or np.sum(mask[y : mask.shape[0], x]) == mask.shape[0] - y - 1
        or np.sum(mask[y, 0:x]) == x
    ):
        return True
    else:
        return False


@decorator_util.jit()
def total_border_pixels_from_mask_and_edge_pixels(
    mask, edge_pixels, mask_1d_index_to_mask_2d_index
):
    """Compute the total number of borders-pixels in a mask."""

    border_pixel_total = 0

    for i in range(edge_pixels.shape[0]):

        if check_if_border_pixel(mask, edge_pixels[i], mask_1d_index_to_mask_2d_index):
            border_pixel_total += 1

    return border_pixel_total


@decorator_util.jit()
def border_1d_indexes_from_mask(mask):
    """Compute a 1D array listing all borders pixel indexes in the mask. A borders pixel is a pixel which:

     1) is not fully surrounding by False mask values.
     2) Can reach the edge of the array without hitting a masked pixel in one of four directions (upwards, downwards,
     left, right).

     The borders pixels are thus pixels which are on the exterior edge of the mask. For example, the inner ring of edge \
     pixels in an annular mask are edge pixels but not borders pixels."""

    edge_pixels = edge_1d_indexes_from_mask(mask=mask)
    mask_1d_index_to_mask_2d_index = mask_mapping_util.sub_mask_1d_index_to_sub_mask_2d_index_from_mask_and_sub_grid_size(
        mask=mask, sub_grid_size=1
    )

    border_pixel_total = total_border_pixels_from_mask_and_edge_pixels(
        mask=mask,
        edge_pixels=edge_pixels,
        mask_1d_index_to_mask_2d_index=mask_1d_index_to_mask_2d_index,
    )

    border_pixels = np.zeros(border_pixel_total)

    border_pixel_index = 0

    for edge_pixel_index in range(edge_pixels.shape[0]):

        if check_if_border_pixel(
            mask=mask,
            edge_pixel_1d=edge_pixels[edge_pixel_index],
            mask_1d_index_to_mask_2d_index=mask_1d_index_to_mask_2d_index,
        ):

            border_pixels[border_pixel_index] = edge_pixels[edge_pixel_index]
            border_pixel_index += 1

    return border_pixels


def sub_border_pixel_1d_indexes_from_mask_and_sub_grid_size(mask, sub_grid_size):
    """Compute a 1D array listing all borders pixel indexes in the mask. A borders pixel is a pixel which:

     1) is not fully surrounding by False mask values.
     2) Can reach the edge of the array without hitting a masked pixel in one of four directions (upwards, downwards,
     left, right).

     The borders pixels are thus pixels which are on the exterior edge of the mask. For example, the inner ring of edge \
     pixels in an annular mask are edge pixels but not borders pixels."""

    border_pixels = border_1d_indexes_from_mask(mask=mask)

    sub_border_pixels = np.zeros(shape=border_pixels.shape[0])

    mask_1d_index_to_sub_mask_2d_indexes = mask_mapping_util.mask_1d_index_to_sub_mask_1d_indexes_from_mask(
        mask=mask, sub_grid_size=sub_grid_size
    )

    masked_sub_grid_1d = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
        mask=mask,
        pixel_scales=(1.0, 1.0),
        sub_grid_size=sub_grid_size,
        origin=(0.0, 0.0),
    )
    mask_centre = grid_util.grid_centre_from_grid_1d(grid_1d=masked_sub_grid_1d)

    for (border_1d_index, border_pixel) in enumerate(border_pixels):
        sub_border_pixels_of_border_pixel = mask_1d_index_to_sub_mask_2d_indexes[
            int(border_pixel)
        ]

        sub_border_pixels[
            border_1d_index
        ] = grid_util.furthest_grid_1d_index_from_grid_1d_indexes_and_centre(
            grid_1d=masked_sub_grid_1d,
            grid_1d_indexes=sub_border_pixels_of_border_pixel,
            coordinate=mask_centre,
        )

    return sub_border_pixels


@decorator_util.jit()
def edge_buffed_mask_from_mask(mask):

    edge_buffed_mask = mask.copy()

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                if y + 1 <= mask.shape[0] - 1:
                    edge_buffed_mask[y + 1, x] = False
                if y - 1 >= 0:
                    edge_buffed_mask[y - 1, x] = False
                if x + 1 <= mask.shape[1] - 1:
                    edge_buffed_mask[y, x + 1] = False
                if x - 1 >= 0:
                    edge_buffed_mask[y, x - 1] = False
                if y + 1 <= mask.shape[0] - 1 and x + 1 <= mask.shape[1] - 1:
                    edge_buffed_mask[y + 1, x + 1] = False
                if y + 1 <= mask.shape[0] - 1 and x - 1 >= 0:
                    edge_buffed_mask[y + 1, x - 1] = False
                if y - 1 >= 0 and x + 1 <= mask.shape[1] - 1:
                    edge_buffed_mask[y - 1, x + 1] = False
                if y - 1 >= 0 and x >= 0:
                    edge_buffed_mask[y - 1, x - 1] = False

    return edge_buffed_mask


def rescaled_mask_2d_from_mask_2d_and_rescale_factor(mask_2d, rescale_factor):
    rescaled_mask = rescale(
        image=mask_2d,
        scale=rescale_factor,
        mode="edge",
        anti_aliasing=False,
        multichannel=False,
    )

    rescaled_mask[0, :] = 1
    rescaled_mask[rescaled_mask.shape[0] - 1, :] = 1
    rescaled_mask[:, 0] = 1
    rescaled_mask[:, rescaled_mask.shape[1] - 1] = 1
    return np.isclose(rescaled_mask, 1)
